#!/usr/bin/env python3
""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import sys,os,psutil,argparse
import shlex
import json
import glob
import signal
if os.getuid() != 0:
    print("This script must be run as 'root'")
    sys.exit(1)
sys.path.append("/opt/cyops/scripts/")
from fsr.commonlibs.utils import CustomHelpFormatter, GenericPGEngine, is_forticloud_instance, is_enterprise_instance, is_docker
from csadm.logs.logger_defs import get_logger
from csadm.utils import *
import fsr.fortimonitor.handlers.cli
LOG_FILE = "/var/log/cyops/csadm/system.log"
MODE_TRIAL = 'trial'
MODE_STANDARD = 'standard'
MODE_OPTIMAL = 'optimal'
MODE_PROD = 'prod'
MODE_ALL = [MODE_TRIAL, MODE_STANDARD, MODE_OPTIMAL, MODE_PROD]
EXPOSED_MODE_ALL = [MODE_STANDARD, MODE_OPTIMAL]
ENV_MODE_MAINT = 'maintenance'
ENV_MODE_UPGRADE = 'upgrade'
ENV_MODE_UPGRADING = 'upgrading'
ENV_MODE_OPERATIONAL = 'operational'
ENV_MODE_ALL = [ENV_MODE_MAINT, ENV_MODE_UPGRADE, ENV_MODE_UPGRADING, ENV_MODE_OPERATIONAL]
EXPOSED_ENV_MODE_ALL = [ENV_MODE_UPGRADE, ENV_MODE_OPERATIONAL]
RESOURCE_ES = "elasticsearch"
RESOURCE_PG = "postgres"
RESOURCE_PG_48 = "postgres48"
RESOURCE_TOMCAT = "tomcat"
RESOURCE_PHP = "php"
RESOURCE_PHP_48 = "php48"
RESOURCE_MQ = "mq"
RESOURCE_WORKFLOW = "workflow"
RESOURCE_AUDIT = "audit"
RESOURCE_HARDWARE = "hardware"
DIR_DATA_RESET = "/var/lib/pgsql/.data-reset"
SYSTEM_CONFIG = "/opt/cyops/configs/common/system-config.yml"
OPTIMAL_SYSTEM_CONFIG = "/opt/cyops/configs/integrations/fortisoar_optimal_configuration.yml"
DB_CONF_FILE = "/opt/cyops/configs/database/db_config.yml"

logger = get_logger(__name__, LOG_FILE)
set_logger_for_utils(logger)

config = None
args = None
cpu_core = None
ram = None
services_to_restart = []
######
# add helper functions below this
#########

def check_positive_integer(i):
    i = int(i)
    if i <= 0:
        raise argparse.ArgumentTypeError("%s needs to be a positive integer" % i)
    return i

def signal_handler(signal, frame):
    sys.exit(1)

signal.signal(signal.SIGINT, signal_handler)

def populate_config(config_path=None):
    if config_path is None:
        config_path = SYSTEM_CONFIG
    logger.info(f"Loading configuration from the file [{config_path}].")
    global config
    config = read_yaml(config_path)

def get_config(resource):
    mode = args.mode
    if mode not in config.keys():
        return None
    if resource not in config[mode].keys():
        return None
    return config[mode][resource]


def is_mode_present_in_config():
    return args.mode in config


def get_percent_diff(required, current):
    change_percent = ((float(required) - current)/required) * 100
    return float('%.2f' % (change_percent))

def get_cpu_core_count():
    global cpu_core
    if cpu_core:
        return cpu_core
    cpu_core = int(os.cpu_count())
    return cpu_core

def get_ram():
    global ram
    if ram:
        return ram
    physical_memory = psutil.virtual_memory()
    physical_memory_details = dict(physical_memory._asdict())
    in_gb = float(physical_memory_details['total'])/(1024*1024*1024)
    ram = float('%.2f' % (in_gb))
    return ram

def postgres_version():
    cmd = "psql --version | egrep -o '[0-9]{1,}\.' | cut -d'.' -f1"
    ret, out, err = execute_cmd_with_shell(cmd)
    return out

def restart_services():
    global services_to_restart
    if len(services_to_restart) == 0:
        return
    serv_list = ' '.join(services_to_restart)
    cmd = "systemctl restart " + serv_list
    logger.info("Restarting services [" + serv_list + "]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to restart services with error [" + str(err) + "]")

def enable_celery_autoscale():
    celery_conf = "/etc/celery/celeryd.conf"
    cmd = "sed -i \"s/^\s*#\\+\s*\(CELERYD_OPTS\s*=.*autoscale.*$\)/\\1/g\" " + celery_conf
    logger.info("Enabling the celery autoscaling")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to enable the celery autoscaling with error [" + str(err) + "]")

def set_celery_autoscale_param(max, min):
    celery_conf = "/etc/celery/celeryd.conf"
    cmd = f'sed -i \'s/--autoscale*=*[0-9]\+,[0-9]\+/--autoscale={max},{min}/\' {celery_conf}'
    logger.info("Setting the autoscaling parameters for the celery")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to set the celery autoscaling params with error [" + str(err) + "]")

def disable_celery_autoscale():
    celery_conf = "/etc/celery/celeryd.conf"
    cmd = "sed -i \"s/^\s*CELERYD_OPTS\s*=.*autoscale/#&/g\" " + celery_conf
    logger.info("Disabling the celery autoscaling")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to disable the celery autoscaling with error [" + str(err) + "]")

def set_sealab_process_count(process):
    sealab_conf = "/etc/uwsgi.d/sealab_wsgi.ini"
    cmd = f'sed -i --follow-symlinks \'s/^processes *= *[0-9]\\+/processes = {process}/\' {sealab_conf}'
    logger.info("Setting the process count for the workflow")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to set the process count for the workflow with error [" + str(err) + "]")

def set_playbook_cleanup_mechanism(advance_cleanup):
    sealab_conf = "/opt/cyops-workflow/sealab/sealab/config.ini"
    cleanup_value = 'true' if advance_cleanup else 'false'
    cmd = f"sed -i 's/USE_PG_SQUEEZE: .*/USE_PG_SQUEEZE: {cleanup_value}/' {sealab_conf}"
    logger.info("Configuring the squeeze and repack for the workflow")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure the squeeze for the workflow with error [" + str(err) + "]")
    cmd = f"sed -i 's/USE_PG_REPACK: .*/USE_PG_REPACK: {cleanup_value}/' {sealab_conf}"
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure the repack for the workflow with error [" + str(err) + "]")

def set_playbook_retention_period(pb_retention_ttl):
    logger.info("Configuring the retention period of playbook's log for the workflow")
    response = call_fsr_api('api/3/system_settings/845c05cc-05b3-450e-9afb-df6b6e436321?format=json', 'GET')
    if response['publicValues'].get('workflow_ttl_config'):
        response['publicValues']['workflow_ttl_config'].update({'ttl': pb_retention_ttl})
    call_fsr_api('api/3/system_settings/845c05cc-05b3-450e-9afb-df6b6e436321?format=json', 'PUT', response)

def set_rule_engine_workers_count(workers):
    rule_engine_conf = "/etc/fsr-workflow/fsr-workflow.conf.py"
    cmd = f'sed -i \'s/^workers *= *[0-9]\\+/workers = {workers}/\' {rule_engine_conf}'
    logger.info("Setting the process count for the fsr-workflow")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to set the process count for the fsr-workflow with error [" + str(err) + "]")

def set_integrations_process_count(process):
    integration_conf = "/etc/uwsgi.d/integrations_wsgi.ini"
    cmd = f'sed -i --follow-symlinks \'s/^processes *= *[0-9]\\+/processes = {process}/\' {integration_conf}'
    logger.info("Setting the process count for the integration")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to set the process count for the integration with error [" + str(err) + "]")

def set_php_apcu_memory(acpu_memory_size_in_mb):
    conf_php = "/etc/php.d/40-apcu.ini"
    logger.info("Configuring the PHP APCu memory [" + str(acpu_memory_size_in_mb) + "MB]")
    cmd = "sed -i \"s/apc.shm_size\s*=.*$/apc.shm_size = \\\"" + str(acpu_memory_size_in_mb) + "M\\\"/\" " + conf_php
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure the PHP APCu memory with error [" + str(err) + "]")

def update_cyops_api_conf(max_children, start_servers, min_spare_servers, max_spare_servers, admin_memory_limit_in_mb):
    conf_php = "/etc/php-fpm.d/cyops-api.conf"
    logger.info("Configuring the cyops-api settings")
    cmd = f'sed -i \'s/^pm.max_children\s*=.*$/pm.max_children = {max_children}/\' {conf_php}'
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure php max_children with error [" + str(err) + "]")
    cmd = f'sed -i \'s/^pm.start_servers\s*=.*$/pm.start_servers = {start_servers}/\' {conf_php}'
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure php start_servers with error [" + str(err) + "]")
    cmd = f'sed -i \'s/^pm.min_spare_servers\s*=.*$/pm.min_spare_servers = {min_spare_servers}/\' {conf_php}'
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure php min_spare_servers with error [" + str(err) + "]")
    cmd = f'sed -i \'s/^pm.max_spare_servers\s*=.*$/pm.max_spare_servers = {max_spare_servers}/\' {conf_php}'
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure php max_spare_servers with error [" + str(err) + "]")
    cmd = "sed -i 's/^php_admin_value\[memory_limit\]\s*=\s*[0-9]\+M/php_admin_value[memory_limit] = " + str(admin_memory_limit_in_mb) + "M/' " + conf_php
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure php_admin_value[memory_limit] with error [" + str(err) + "]")

def set_elastic_search_memory_config(xms_size_in_gb, xmx_size_in_gb):
    jvm_options_es = "/etc/elasticsearch/jvm.options.d/fsr.options"
    logger.info("Configuring the Elasticsearch Xms memory [" + str(xms_size_in_gb) + "GB]")
    cmd = "sed -i \"s/^-Xms.*$/-Xms" + str(xms_size_in_gb) + "g/\" " + jvm_options_es
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xms value for the Elasticsearch with error [" + str(err) + "]")
    cmd = "sed -i \"s/^-Xmx.*$/-Xmx" + str(xmx_size_in_gb) + "g/\" " + jvm_options_es
    logger.info("Configuring the Elasticsearch Xmx memory [" + str(xmx_size_in_gb) + "GB]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xmx value for the Elasticsearch with error [" + str(err) + "]")

def mask_elastic_search():
    stop_es = 'systemctl stop cyops-search elasticsearch'
    disable_es = 'systemctl disable cyops-search elasticsearch'
    mask_es = 'systemctl mask cyops-search elasticsearch'
    ret, out, err = execute_cmd(stop_es, True)
    if ret != 0:
        print_error("Failed to stop elastic search with error [" + str(err) + "]")
    ret, out, err = execute_cmd(disable_es, True)
    if ret != 0:
        print_error("Failed to disable elastic search with error [" + str(err) + "]")
    ret, out, err = execute_cmd(mask_es, True)
    if ret != 0:
        print_error("Failed to mask elastic search with error [" + str(err) + "]")

def unmask_elastic_search():
    start_es = 'systemctl start elasticsearch cyops-search'
    enable_es = 'systemctl enable elasticsearch cyops-search'
    unmask_es = 'systemctl unmask elasticsearch cyops-search'
    ret, out, err = execute_cmd(unmask_es, True)
    if ret != 0:
        print_error("Failed to unmask elastic search with error [" + str(err) + "]")
    ret, out, err = execute_cmd(enable_es, True)
    if ret != 0:
        print_error("Failed to enable elastic search with error [" + str(err) + "]")
    ret, out, err = execute_cmd(start_es, True)
    if ret != 0:
        print_error("Failed to start elastic search with error [" + str(err) + "]")

def stop_mask_celeryd():
    stop_celeryd = 'systemctl stop celeryd'
    mask_celeryd = 'systemctl mask celeryd'
    ret, out, err = execute_cmd(stop_celeryd, True)
    if ret != 0:
        print_error("Failed to stop celeryd service with error [" + str(err) + "]")
    ret, out, err = execute_cmd(mask_celeryd, True)
    if ret != 0:
        print_error("Failed to mask celeryd service with error [" + str(err) + "]")

def unmask_start_celeryd():
    start_celeryd = 'systemctl start celeryd'
    unmask_celeryd = 'systemctl unmask celeryd'
    ret, out, err = execute_cmd(unmask_celeryd, True)
    if ret != 0:
        print_error("Failed to unmask celeryd service with error [" + str(err) + "]")
    ret, out, err = execute_cmd(start_celeryd, True)
    if ret != 0:
        print_error("Failed to start celeryd service with error [" + str(err) + "]")

def call_fsr_api(endpoint, method='GET', payload=None, is_das=False):
    if is_das:
        url = f'https://localhost:8443/{endpoint}'
    else:
        url = f'https://localhost/{endpoint}'
    cmd = f'/opt/cyops/scripts/api_caller.py --endpoint "{url}" --method {method}'
    if payload is not None:
        cmd += f' --payload {shlex.quote(json.dumps(payload))}'
    if is_das:
        cmd += " --das"
    cmd += ' --silent'

    ret, output, err = execute_cmd_with_shell(cmd, True)
    if ret == 0:
        return json.loads(output) if output else {}
    else:
        if output is not None:
            try:
                return json.loads(output)
            except Exception as e:
                pass
        print(f"API {url} call failed with error: {err}")
    return None

######
# add helper functions above this
#########

def configure_es():
    config_es = get_config(RESOURCE_ES)
    if config_es is None:
        return
    xms_size_in_gb = int(config_es.get('xms')) if config_es.get('xms') else None
    xmx_size_in_gb = int(config_es.get('xmx')) if config_es.get('xmx') else None
    is_mask = config_es.get('mask')
    if is_mask:
        mask_elastic_search()
    else:
        unmask_elastic_search()
    if (not is_mask) and xms_size_in_gb and xmx_size_in_gb:
        set_elastic_search_memory_config(xms_size_in_gb, xmx_size_in_gb)
    if not is_mask:
        global services_to_restart
        services_to_restart.append("elasticsearch")

def configure_pg():
    ram = get_ram()
    cpu_core = get_cpu_core_count()
    pg_resource = RESOURCE_PG_48 if args.mode != MODE_OPTIMAL and ram > 32 and cpu_core > 8 else RESOURCE_PG
    config_pg = get_config(pg_resource)
    if config_pg is None:
        return
    PG_VERSION = postgres_version()
    max_worker_processes = int(config_pg.get('max_worker_processes')) if config_pg.get(
        'max_worker_processes') else None
    max_parallel_workers_per_gather = int(config_pg.get('max_parallel_workers_per_gather')) if config_pg.get(
        'max_parallel_workers_per_gather') else None
    max_parallel_maintenance_workers = int(config_pg.get('max_parallel_maintenance_workers')) if config_pg.get(
        'max_parallel_maintenance_workers') else None
    max_parallel_workers = int(config_pg.get('max_parallel_workers')) if config_pg.get(
        'max_parallel_workers') else None
    max_connections = int(config_pg.get('max_connections')) if config_pg.get(
        'max_connections') else None
    checkpoint_completion_target = float(config_pg.get('checkpoint_completion_target')) if config_pg.get(
        'checkpoint_completion_target') else None
    default_statistics_target = int(config_pg.get('default_statistics_target')) if config_pg.get(
        'default_statistics_target') else None
    random_page_cost = float(config_pg.get('random_page_cost')) if config_pg.get(
        'random_page_cost') else None
    effective_io_concurrency = int(config_pg.get('effective_io_concurrency')) if config_pg.get(
        'random_page_cost') else None
    huge_pages = config_pg.get('huge_pages')
    # Check if it's explicitly False (due to PyYAML interpretation of 'off')
    if huge_pages is False:
        huge_pages = 'off'
    elif huge_pages is None:
        huge_pages = None
    else:
        huge_pages = str(huge_pages)
    min_wal_size = int(config_pg.get('min_wal_size')) if config_pg.get(
        'min_wal_size') else None
    max_wal_size = int(config_pg.get('max_wal_size')) if config_pg.get(
        'max_wal_size') else None
    effective_cache_size = int(config_pg.get('effective_cache_size')) if config_pg.get(
        'effective_cache_size') else None
    shared_buffer = int(config_pg['shared_buffer'])
    work_mem = int(config_pg['work_mem'])
    maintenance_work_mem = int(config_pg.get('maintenance_work_mem')) if config_pg.get(
        'maintenance_work_mem') else None
    wal_buffers = int(config_pg.get('wal_buffers')) if config_pg.get(
        'wal_buffers') else None

    keys = {
        "max_worker_processes": max_worker_processes,
        "max_parallel_workers_per_gather": max_parallel_workers_per_gather,
        "max_parallel_maintenance_workers": max_parallel_maintenance_workers,
        "max_parallel_workers": max_parallel_workers,
        "max_connections": max_connections,
        "checkpoint_completion_target": checkpoint_completion_target,
        "default_statistics_target": default_statistics_target,
        "random_page_cost": random_page_cost,
        "effective_io_concurrency": effective_io_concurrency,
        "huge_pages": huge_pages,
        # Below values are in GB
        "min_wal_size": f'{min_wal_size}GB',
        "max_wal_size": f'{max_wal_size}GB',
        "effective_cache_size": f'{effective_cache_size}GB',
        # Below values are in MB
        "shared_buffers": f'{shared_buffer}MB',
        "work_mem": f'{work_mem}MB',
        "maintenance_work_mem": f'{maintenance_work_mem}MB',
        "wal_buffers": wal_buffers if wal_buffers == -1 else f'{wal_buffers}MB',
    }

    engine = GenericPGEngine(db_name='sealab', logger=logger)
    try:
        for key, value in keys.items():
            if value:
                logger.info(f"Configuring the PostgreSQL {key} [{value}]")
                result = engine.execute_text_query(f"ALTER SYSTEM SET {key} = '{value}'")
                if result:
                    print_error(f"Failed to configure the PostgreSQL {key} with error [ {str(result)} ]")
    finally:
        engine.dispose_engine()
    global services_to_restart
    services_to_restart.append("postgresql-" + str(PG_VERSION))

def configure_tomcat():
    config_tomcat = get_config(RESOURCE_TOMCAT)
    if config_tomcat is None:
        return
    xms_size_in_mb = int(config_tomcat['xms'])
    xmx_size_in_mb = int(config_tomcat['xmx'])
    conf_tomcat = "/opt/cyops-tomcat/bin/setenv.sh"
    logger.info("Configuring the Tomcat Xms memory [" + str(xms_size_in_mb) + "MB]")
    cmd = "sed -i \"s/-Xms.*M/-Xms" + str(xms_size_in_mb) + "M/\" " + conf_tomcat
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xms value for the Tomcat with error [" + str(err) + "]")
    cmd = "sed -i \"s/-Xmx.*M/-Xmx" + str(xmx_size_in_mb) + "M/\" " + conf_tomcat
    logger.info("Configuring the Tomcat Xmx memory [" + str(xmx_size_in_mb) + "MB]")
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to configure Xmx value for the Tomcat with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("cyops-tomcat")

def configure_php():
    php_resource = RESOURCE_PHP_48 if args.mode != MODE_OPTIMAL and get_ram() > 32 and get_cpu_core_count() > 8 else RESOURCE_PHP
    config_php = get_config(php_resource)
    if config_php is None:
        return
    acpu_memory_size_in_mb = config_php.get('acpu')
    if acpu_memory_size_in_mb:
        acpu_memory_size_in_mb = int(acpu_memory_size_in_mb)
    admin_memory_limit_in_mb = config_php.get('admin_memory_limit')
    if admin_memory_limit_in_mb:
        admin_memory_limit_in_mb = int(admin_memory_limit_in_mb)
    max_children = config_php.get('max_children')
    start_servers = config_php.get('start_servers')
    min_spare_servers = config_php.get('min_spare_servers')
    max_spare_servers = config_php.get('max_spare_servers')
    if (acpu_memory_size_in_mb is None and admin_memory_limit_in_mb is None
            and max_children is None and start_servers is None
            and min_spare_servers is None and max_spare_servers is None):
        return
    if acpu_memory_size_in_mb:
        set_php_apcu_memory(acpu_memory_size_in_mb)
    update_cyops_api_conf(max_children, start_servers, min_spare_servers, max_spare_servers, admin_memory_limit_in_mb)
    global services_to_restart
    services_to_restart.append("php-fpm")

def configure_mq(number_of_scheduler=4):
    config_mq = get_config(RESOURCE_MQ)
    if config_mq is None:
        number_of_scheduler = min(6, max(1, get_cpu_core_count() // 2))
    else:
        number_of_scheduler = int(config_mq['scheduler'])
    conf_mq = "/etc/rabbitmq/rabbitmq-env.conf"
    config_key = "RABBITMQ_SERVER_ADDITIONAL_ERL_ARGS"
    cmd = "sed -i '/" + config_key + "/d' " + conf_mq
    ret, out, err = execute_cmd(cmd, True)
    if ret != 0:
        print_error("Failed to remove the existing value for the RabbitMQ scheduler")
        return
    # add empty line first. This ensure that newline character gets added, before we write any configuration.
    cmd = "echo '' >> " + conf_mq
    execute_cmd_with_shell(cmd, True)
    # RABBITMQ_SERVER_ADDITIONAL_ERL_ARGS="+S 4:4"
    cmd = "echo '" + config_key + "=\"+S " + str(number_of_scheduler) + ":" + str(number_of_scheduler) + "\"' >> " + conf_mq
    logger.info("Configuring the RabbitMQ scheduler [" +str(number_of_scheduler) + "]")
    ret, out, err = execute_cmd_with_shell(cmd, True)
    if ret != 0:
        print_error("Failed to configure the RabbitMQ scheduler with error [" + str(err) + "]")
    global services_to_restart
    services_to_restart.append("rabbitmq-server")

def configure_workflow():
    config_workflow = get_config(RESOURCE_WORKFLOW)
    if config_workflow is None:
        return
    autoscale = config_workflow.get('autoscale')
    max_consumer = config_workflow.get('max_consumer')
    min_consumer = config_workflow.get('min_consumer')
    process = config_workflow.get('process')
    advance_cleanup = config_workflow.get('advance_cleanup')
    pb_retention_ttl = config_workflow.get('pb_retention_ttl')
    rule_engine_worker = config_workflow.get('rule_engine_worker')
    if (autoscale is None and max_consumer is None
            and min_consumer is None and process is None):
        return

    if autoscale:
        enable_celery_autoscale()
        if min_consumer is None or max_consumer is None:
            min_consumer = get_cpu_core_count()
            max_consumer = min_consumer + 4
        set_celery_autoscale_param(max_consumer, min_consumer)
    else:
        disable_celery_autoscale()
    global services_to_restart
    services_to_restart.append("celeryd")
    if process:
        if args.mode != MODE_OPTIMAL and get_cpu_core_count() > 10:
            process = get_cpu_core_count()
        set_sealab_process_count(process)
        set_integrations_process_count(process)
    if advance_cleanup is not None:
        set_playbook_cleanup_mechanism(advance_cleanup)
    if pb_retention_ttl:
        set_playbook_retention_period(pb_retention_ttl)
    if rule_engine_worker:
        set_rule_engine_workers_count(rule_engine_worker)
    if process is not None or rule_engine_worker is not None or advance_cleanup is not None:
        services_to_restart.append("uwsgi")

def configure_audit():
    config_audit = get_config(RESOURCE_AUDIT)
    if config_audit is None:
        return
    audit_log_level = config_audit.get('audit_log_level', 'Detailed')
    if audit_log_level in ['Detailed', 'Restricted']:
        payload = {
            "section": "AUDIT_CONFIG",
            "key": "audit_log_level",
            "value": audit_log_level,
            "dataType": "STRING"
        }
        logger.info(f"Configuring the audit log level [{audit_log_level}]")
        call_fsr_api('api/gateway/config/config', 'PUT', payload)

def validate_resource(required_cpu=8, required_ram=16, ram_percent_error_margin=10.00):
    config_hardware = get_config(RESOURCE_HARDWARE)
    if config_hardware is not None:
        required_cpu = config_hardware['cpu']
        required_ram = config_hardware['ram']
        ram_percent_error_margin = config_hardware['ram_percent_error_margin']
    current_cpu = get_cpu_core_count()
    current_ram = get_ram()
    logger.info("Available CPU core [" + str(current_cpu) + "]")
    logger.info("Available RAM [" + str(current_ram) + "GB]")
    error = ""
    if current_cpu < required_cpu:
        error = error + "Ensure you have [" + str(required_cpu) + "] CPU core" + "\n"
    ram_percent_diff = get_percent_diff(required_ram, current_ram)
    # This will work if RAM is 3.6GB when checked for 4GB RAM validation
    # On FortiCloud Aailable RAM [31.26GB], percent diff is 2.31
    if ram_percent_diff > ram_percent_error_margin:
        error = error + "Ensure you have [" + str(required_ram) + "GB] RAM" + "\n"
    if error != "":
        print_error(error)
        sys.exit(1)

def perform_hardening():
    global args
    logger.info("Configuring " +  str(args.mode) + " mode")
    if not args.is_force:
        validate_resource()
    else:
        logger.info("'--force' is opted, skipping resource validation")
    configure_workflow()
    configure_php()
    configure_tomcat()
    configure_mq()
    configure_pg()
    configure_es()
    configure_audit()
    if not args.skip_service_restart:
        restart_services()

def run_data_reset():
    backup_files = glob.glob(DIR_DATA_RESET + '/DR_BACKUP_*.tgz')
    if (backup_files is None) or len(backup_files) == 0:
        print_error("Backup data is not found for the data reset operation. Therefore, data reset cannot be performed.")
        sys.exit(1)
    msg =       "WARNING:\n"
    msg = msg + "Resetting data causes data loss, i.e., this operation will reset data to the point\n"
    msg = msg + "where you first saw the EULA on the console. This means that all your connector\n"
    msg = msg + "configurations, created users, created playbooks, playbook execution logs, audit\n"
    msg = msg + "logs, etc. will be lost."
    response = take_choice_input(msg=msg)
    logger.debug("User entered [" + response + "]")
    if response.lower() != 'yes':
        sys.exit(0)
    cmd = "csadm db --restore " + str(backup_files[0]) + " --no-interaction --reset-data"
    ret = os.system(cmd)
    exit_code = 0
    if ret != 0:
        exit_code = 1
    sys.exit(exit_code)

def disk_management():
    s_args = ' '.join(sys.argv[2:])
    cmd_disk = "/opt/cyops/scripts/csadm/commands/disk.sh --" + s_args
    ret = os.system(cmd_disk)
    exit_code = 0
    if ret != 0:
        exit_code = 1
    sys.exit(exit_code)

def get_device_uuid():
    cmd = "csadm license --get-device-uuid"
    ret, out, err = execute_cmd(cmd, False)
    if ret != 0:
        logger.error("Failed to get device UUID with error [" + str(err) + "]")
    return out

def is_passive_node():
    json_path = "/opt/cyops-auth/.list_nodes.json"
    nodes = None
    isPassive = False
    nodeId = get_device_uuid()

    try:
        with open(json_path, 'r') as json_contents:
            nodes = json.load(json_contents)
    except Exception as e:
        logger.error(str(e))

    if nodes is not None and nodeId is not None:
        for node in nodes:
            if node['nodeId'] == nodeId and node['status'] == "passive":
                isPassive = True
                break
    
    return isPassive

#
# For env CLI commands database changes are handle as a part of DAS module.
# Commands are run on SELF node.
# 'csadm system env --mode <upgrade, maintenance, operational>' command set env mode of the self node
# 'csadm system env show' command fetch current environment mode of the self node
#
# Handling of 'csadm system env --mode <upgrade, maintenance, operational>' command
#
def handle_env():
    global args
    env_mode = args.env_mode
    if env_mode:
        payload = {"mode": env_mode}
        result = call_fsr_api('env', 'PUT', payload, is_das=True)
        if result is not None and str(result.get("response")) == "Success":
            # Get node status and check for Passive node.
            isPassive = is_passive_node()
            print("Environment mode for the node is set to {}".format(env_mode))
            if env_mode in [ENV_MODE_UPGRADE,ENV_MODE_UPGRADING] and not isPassive:
                # Stop and mask celeryd service for upgrade mode
                print("Stopping and masking celeryd service")
                stop_mask_celeryd()
            elif env_mode == ENV_MODE_OPERATIONAL and not isPassive:
                # Unmask and start celeryd service for operational mode
                print("Unmasking and starting celeryd service")
                unmask_start_celeryd()
        elif result is not None and result.get("Error") is not None:
            print(f"Error: {result.get('Error')}")
        else:
            print("Failed to set the environemnt mode for the node")

#
# Handling of 'csadm system env show' command
#
def handle_env_show():
    result = call_fsr_api('env', 'GET', is_das=True)
    if result is not None:
        print("Current environment mode of the node is: {}".format(result.get("mode")))
    else:
        print("Failed ot show environment mode for the node")
        
def define_parser_subcmd_data_reset(parser, subparsers):
    parser_data_reset = subparsers.add_parser('data-reset')
    parser_data_reset.set_defaults(operation='data-reset')

def define_parser_subcmd_config(parser, subparsers):
    parser_config = subparsers.add_parser('config', help='System configuration')

    parser_config.set_defaults(operation='config')
    parser_config.add_argument('--mode',
                               help=f"System mode (choices: {', '.join(EXPOSED_MODE_ALL)})",
                               type=validate_mode,  # Use custom validation
                               dest='mode',
                               required=True)

    parser_config.add_argument('--skip-service-restart',
                               help=argparse.SUPPRESS,
                               dest='skip_service_restart',
                               action='store_true',
                               required=False)

    parser_config.add_argument('--force',
                               help=argparse.SUPPRESS,
                               dest='is_force',
                               action='store_true',
                               required=False)

    parser_config.add_argument('--path',
                               help=f'Configuration file path (only available in {MODE_OPTIMAL} mode)',
                               dest='path',
                               required=False)

# Custom type to validate input
def validate_mode(mode):
    if mode not in MODE_ALL:
        raise argparse.ArgumentTypeError(f"Invalid mode: '{mode}'. Must be one of {EXPOSED_MODE_ALL}.")
    return mode

# Create argument parser for disk command
def define_parser_subcmd_disk(subparsers):
    # Help of disk action is displayed using parser_disk.print_help, if user enters just "csadm system disk"
    # Need to set it a global variable for that
    global parser_disk
    parser_disk = subparsers.add_parser('disk', help='Disk Management')
    parser_disk.set_defaults(operation='disk')

    visible_options = "{expand-lv}"
    sub_parser_disk = parser_disk.add_subparsers(metavar=visible_options)

    sub_parser_disk_expand_lv = sub_parser_disk.add_parser("expand-lv", help="Expand the logical volume.")
    sub_parser_disk_expand_lv.set_defaults(operation='expand-lv')

    # Command to get list of logical volumes
    # Expansion of swap and root partition is not supported. It should be handled manually
    cmd_get_logical_volume_list = "lvs -o lv_name --noheadings | grep -vw swap | grep -vw root | sed 's/ *//g' | tr '\n' ' ' | sed 's/.$//' "
    stream = os.popen(cmd_get_logical_volume_list)
    cmd_op_get_logical_volume_list = stream.read()
    l_logical_volumes = []
    l_logical_volumes = cmd_op_get_logical_volume_list.split(' ')
    stream.close()

    # Required parameter for expand-lv
    sub_parser_disk_expand_lv.add_argument('--logical-volume',
                                           help='Name of the logical volume to expand. Expansion of swap and root logical volumes is not supported.',
                                           choices=l_logical_volumes,
                                           dest='logical_volume',
                                           required=True)

    # --disk and --use-vg cannot be used together.
    # Either of them is required though
    group_sub_parser_disk_expand_lv = sub_parser_disk_expand_lv.add_mutually_exclusive_group(required=True)

    # Scan for new disks
    d_scsi_host="/sys/class/scsi_host"
    for s_host in os.listdir(d_scsi_host):
        f_host_full_path=d_scsi_host + "/" + s_host
        s_cmd="echo '- - -' >>" + f_host_full_path + "/scan"
        os.system(s_cmd)

    # Get the list of disks
    cmd_get_disk_list = "lsblk -o KNAME,TYPE | grep disk | awk '{print $1}' | tr '\n' ' ' | sed 's/.$//'"
    stream = os.popen(cmd_get_disk_list)
    cmd_op_get_disk_list = stream.read()
    l_disks = []
    l_disks = cmd_op_get_disk_list.split(' ')
    stream.close()

    # Either of single disk name or "all" are possible choices of this parameter
    group_sub_parser_disk_expand_lv.add_argument('--disk',
                                           help="Disk name to expand the logical volume.",
                                           choices=l_disks,
                                           dest='use_disk',
                                           required=False)

    # --use-vg is a flag
    group_sub_parser_disk_expand_lv.add_argument('--use-vg',
                                           help="Consume the available free disk space from the volume group this logical volume is part of.",
                                           dest='use_vg',
                                           action='store_true',
                                           required=False)

    # If --size is not mentioned, use 100% of either disk or vg
    sub_parser_disk_expand_lv.add_argument('--size',
                                           help="Size in gigabytes (GB), to be consumed from the disk or volume group this logical volume is part of. Omit this option for 100 percent expansion from the disk or the volume group.",
                                           type=check_positive_integer,
                                           required=False)
                                        
    sub_parser_disk_expand_lv.add_argument('--validate',
                                           help="Validates the inputs passed for csadm system disk expand-lv command.",
                                           dest='validate',
                                           action='store_true',
                                           required=False)

def define_parser_subcmd_env(parser, subparsers):
    parser_env = subparsers.add_parser('env', help='HA Upgrade Management')
    parser_env.set_defaults(operation='env')

    parser_env.add_argument('--mode',
                               help=f"System env mode (choices: {', '.join(EXPOSED_ENV_MODE_ALL)})",
                               type=validate_env_mode,  # Use custom validation
                               dest='env_mode',
                               required=False)
    
    subparsers_env = parser_env.add_subparsers()
    parser_env_show = subparsers_env.add_parser('show',
                                                     help="To view system environment details")
    parser_env_show.set_defaults(operation='env-show')

# Custom type to validate env mode
def validate_env_mode(mode):
    if mode not in ENV_MODE_ALL:
        raise argparse.ArgumentTypeError(f"Invalid env mode: '{mode}'. It must be one of the following: {EXPOSED_ENV_MODE_ALL}.")
    return mode

def define_parser_subcmd_fortimonitor(parser, subparsers):
    fsr.fortimonitor.handlers.cli.define_parser(parser, subparsers)

def define_parser(parser, subparsers):
    if is_enterprise_instance():
        define_parser_subcmd_config(parser, subparsers)
    if is_forticloud_instance():
        define_parser_subcmd_data_reset(parser, subparsers)
    # Disk option is not applicable for docker instance
    if not is_docker():
        define_parser_subcmd_disk(subparsers)
    define_parser_subcmd_env(parser, subparsers)
    define_parser_subcmd_fortimonitor(parser, subparsers)

# TBD: to display example commands for disk

def start():
    parser = argparse.ArgumentParser(prog="csadm system", formatter_class=CustomHelpFormatter)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'

    # Disk option is not applicable for docker instance
    if is_docker():
        visible_options = "{env, fortimonitor}"
        subparsers = parser.add_subparsers(metavar=visible_options)
    else:
        visible_options = "{config, disk, env, fortimonitor}"
        subparsers = parser.add_subparsers(metavar=visible_options)

    define_parser(parser, subparsers)

    global args
    args = parser.parse_args()

    if not 'operation' in dir(args):
        parser.print_help(sys.stderr)
        sys.exit(1)
    if str(args.operation) == "config":
        config_path = None
        if args.mode == MODE_PROD:
            args.mode = MODE_STANDARD
            logger.info(f"WARN: 'prod' mode has been deprecated.")
        if args.mode in {MODE_TRIAL, MODE_STANDARD}:
            config_path = SYSTEM_CONFIG
            if args.path:
                logger.info(f"WARN: Ignoring --path value and using the default configuration "
                            f"path as the mode is set to {args.mode}.")
        elif args.mode == MODE_OPTIMAL:
            config_path = args.path if args.path else OPTIMAL_SYSTEM_CONFIG
            if not os.path.exists(config_path):
                if args.path:
                    logger.info(f"{args.path} configuration file is not present.")
                else:
                    logger.info("Required configuration file is not present. Please install "
                                "'FortiSOAR Health Assessment Solution Pack' and generate optimal configurations.")
                sys.exit(1)
        populate_config(config_path)
        if not is_mode_present_in_config():
            logger.info(f"The configuration file [{config_path}] is missing the key '{args.mode}'.")
            sys.exit(1)
        perform_hardening()
        return
    elif str(args.operation) == "disk":
        parser_disk.print_help(sys.stderr)
        sys.exit(1)
    elif str(args.operation) == "expand-lv":
        disk_management()
        return
    elif str(args.operation) == "data-reset":
        run_data_reset()
        return
    elif str(args.operation) == "env":
        handle_env()
        return
    elif str(args.operation) == "env-show":
        handle_env_show()
        return
    elif str(args.operation).startswith("forti-monitor"):
        fsr.fortimonitor.handlers.cli.handle_cli_ops(args)
        return

    # Execution should not come to this stage. If comes, then print help
    parser.print_help(sys.stderr)

if __name__ == '__main__':
    start()
sys.exit(0)
