#!/bin/sh
# Copyright start
# Copyright (C) 2008 - 2024 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end

# The script performs:
# 1. Addition of newly attached disks (if added)
# 2. Extension of the partitions (in of case free disk space)
#   if disk is < 200G, to lvroot logical volume
#   else to lvdata logical volume

set -x



# get logical volume path from lv name
get_logical_volume_path(){
    local s_logical_volume_name=$1
    echo $(lvdisplay -c | grep $s_logical_volume_name | cut -d':' -f1)
}

# get file system mounted on lv 
get_logical_volume_filesystem(){
    local s_logical_volume_name=$1
    echo $(df -kh | grep -w $s_logical_volume_name | awk '{print $6}')
}

# get volume group on which lv is mounted.
get_volume_group(){
    local s_logical_volume_name=$1
    echo $(lvs | grep -w $s_logical_volume_name | awk '{print $2}')
}





init(){
    # pgsql, es and mq are mounted from logical volume: lvdata
    s_logical_volume_data="lvdata"
    s_volume_group_data="vgdata"

    # filesystem    LV          VG
    # /             lvroot      vgos
    # swap          lvswap      vgos
    # /home         lvhome      vgos
    # /temp         lvtmp       vgos
    s_logical_volume_root="lvroot"
    s_logical_volume_swap="lvswap"
    s_logical_volume_home="lvhome"
    s_logical_volume_temp="lvtmp"

    s_volume_group_os="vgos"

    i_free_disk_space_percent_threshold=5

    flag_extend_logical_volume_data=0
    flag_extend_logical_volume_os=0

    # Additional disk sizes 
    i_additional_temp_size=3
    i_additional_home_size=8
    i_additional_swap_size=2

}

add_new_disks(){
    # Overall logic:
    # Run disk scan command
    # Get the list of all disks
    # Loop through the list
    #   its new disk if its entry is not found in the output of pvs command
    # prepare the list of new disks
    # Get name of the disk. Ex: sdi
    # Loop through each disk
    #   partition the disk
    #       parted --script --align optimal /dev/sdi mklabel gpt mkpart primary 0% 100%
    #       o/p: partioned /dev/sdi1
    #   create physical volume
    #       pvcreate /dev/sdi1
    #       o/p: /dev/sdi1
    #       add newly created physical volume to existing volume group
    #           vgextend <vgroup_name> /dev/sdi1
    #           o/p: <vgroup_name> with additional disk space
    #           if the newly created physical volume size is > 200GiB, then it is used for extendending the volume group for data (vgdata)
    #           if the newly created physical volume size is <= 200GiB, then it is used for extendending the volume group for os (vgos)

    # Scan each scsi host device for new disks
    for host in /sys/class/scsi_host/*; do
        echo "- - -" >> $host/scan
    done

    # Get the list of disks
    l_disks=`lsblk -o KNAME,TYPE | \
             grep disk | \
             awk '{print $1}'`

    # Loop through the list
    for s_disk in $l_disks; do
        #   its new disk if its entry is not found in the output of pvs command
        pvs | grep /dev/$s_disk
        if [ $? -ne 0 ]; then
            l_new_disks="$l_new_disks /dev/$s_disk"

            # Set the flag to extend the logical volume
            flag_extend_logical_volume=1
        fi
    done

    # Loop through each disk
    for s_new_disk_device in $l_new_disks ; do
        # Get the disk name from device name
        s_new_disk_name=`echo $s_new_disk_device | sed "s#/dev/\(.*\)#\1#g"`

        # Create the partition
        parted --script \
               --align optimal \
               $s_new_disk_device \
               mklabel gpt \
               mkpart primary 0% 100%

        sleep 5

        # Create the physical volume
        s_new_physical_volume=${s_new_disk_device}1
        pvcreate $s_new_physical_volume

        s_partition_size=$( lsblk -o SIZE $s_new_physical_volume | tail -1 )
        size_in_gb=${s_partition_size%G}

        local s_volume_group="vgos"
        if (( size_in_gb > 200 )); then
            # Attached disk is greater than 200G, hence expand vgdata volume group
            s_volume_group="vgdata"
            flag_extend_logical_volume_data=1
        else
            # Attached disk is less than 200G, hence expand vgos volume group
            s_volume_group="vgos"
            flag_extend_logical_volume_os=1
        fi

        # Add the newly created physical volume to existing volume group
        echo "Extending Volume Group : $s_volume_group on Physical Volume: $s_new_physical_volume"
        vgextend $s_volume_group $s_new_physical_volume

    done
}

extend_existing_partitions(){
    local s_volume_group=$1
    # Get the list of partition names of the volume group on which volume group is mounted


    l_existing_partitions=`vgs $s_volume_group -o +devices | \
                          grep $s_volume_group | \
                          awk '{print $8}' | \
                          cut -d\( -f1 | \
                          uniq`

    for s_existing_partition in $l_existing_partitions; do

        # Get the disk name of $s_existing_partition
        s_disk_name=`echo $s_existing_partition | head -c -2`

        # Get the partition number
        i_partition_number=`echo $s_existing_partition | grep -Eo '[0-9]+$'`

        # Get physical volume name
        s_physical_volume=$s_disk_name$i_partition_number

        # Calculate the free disk percentage on partition
        i_free_disk_space_percent=`parted $s_disk_name unit '%' print free | \
                                            grep 'Free Space' | \
                                            tail -n1 | \
                                            awk '{print $3}' | \
                                            cut -d. -f1`

        # Extend the LVM only if free disk space on the partition is more than set percentage threshold
        if [ $i_free_disk_space_percent -gt $i_free_disk_space_percent_threshold ];then

            # Grow the partition to its maximum available space
            growpart $s_disk_name $i_partition_number

            # Grow the physical volume to fully occupy the partition
            pvresize $s_physical_volume

            # Set the flag to extend the logical volume
            if [ $s_volume_group = "vgos" ]; then
                flag_extend_logical_volume_os=1
            elif [ $s_volume_group = "vgdata" ]; then
                flag_extend_logical_volume_data=1
            fi

        fi
    done
}

# Extend the logical volume to 100% and
# grow the filesystem to 100%


extend_logical_volume_data(){
    local s_logical_volume_path=$(get_logical_volume_path $s_logical_volume_data)
    local s_logical_volume_filesystem=$(get_logical_volume_filesystem $s_logical_volume_data)

    # Extend the logical volume to fully occupy the volume group
    echo "INFO: Extending logical volume $s_logical_volume_path to occupy 100% free space on volume group."
    lvextend --extents +100%FREE $s_logical_volume_path

    # Grow the filesystem to fully occupy the logical volume
    echo "INFO: Growing filesystem $s_logical_volume_filesystem to occupy 100% free space on logical volume."
    xfs_growfs $s_logical_volume_filesystem
}

extend_logical_volume_os(){

    local s_logical_volume_path_root=$(get_logical_volume_path $s_logical_volume_root)
    local s_logical_volume_path_swap=$(get_logical_volume_path $s_logical_volume_swap)
    local s_logical_volume_path_home=$(get_logical_volume_path $s_logical_volume_home)
    local s_logical_volume_path_temp=$(get_logical_volume_path $s_logical_volume_temp)

    local s_logical_volume_filesystem_root=$(get_logical_volume_filesystem $s_logical_volume_root)
    local s_logical_volume_filesystem_swap=$(get_logical_volume_filesystem $s_logical_volume_swap)
    local s_logical_volume_filesystem_home=$(get_logical_volume_filesystem $s_logical_volume_home)
    local s_logical_volume_filesystem_temp=$(get_logical_volume_filesystem $s_logical_volume_temp)

    resize_swap $s_logical_volume_path_swap

    lvextend --size +${i_additional_temp_size}G $s_logical_volume_path_temp
    lvextend --size +${i_additional_home_size}G $s_logical_volume_path_home
    lvextend --extents +100%FREE $s_logical_volume_path_root



    xfs_growfs $s_logical_volume_filesystem_root
    xfs_growfs $s_logical_volume_filesystem_home
    xfs_growfs $s_logical_volume_filesystem_temp

}

resize_swap() {
    local s_logical_volume_path_swap=$1

    swapoff $s_logical_volume_path_swap
    lvextend -L +${i_additional_swap_size}G $s_logical_volume_path_swap
    mkswap $s_logical_volume_path_swap
    swapon $s_logical_volume_path_swap

    echo "Swap resized and re-enabled successfully on $s_logical_volume_path_swap."

}

#######
# Main
#######

init

extend_existing_partitions $s_volume_group_os

extend_existing_partitions $s_volume_group_data

add_new_disks

if [ $flag_extend_logical_volume_data -eq 1 ];then
    extend_logical_volume_data
fi

if [ $flag_extend_logical_volume_os -eq 1 ];then
    extend_logical_volume_os
fi


exit
 